/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/10/22.
//

#include "fggl/data/assimp/module.hpp"
#include "fggl/debug/logging.hpp"
#include "fggl/assets/manager.hpp"
#include "fggl/mesh/mesh.hpp"

#include "../../stb/stb_image.h"
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>

#include <filesystem>

namespace fggl::data::models {

	constexpr auto convert(aiVector3D& vec) -> math::vec3 {
		return {vec.x, vec.y, vec.z};
	}

	constexpr auto convert2(aiVector2D& vec) -> math::vec2 {
		return {vec.x, vec.y};
	}

	constexpr auto convert2(aiVector3D& vec) -> math::vec2 {
		return {vec.x, vec.y};
	}

	constexpr auto convert(aiColor3D& col) -> math::vec3 {
		return {col.r, col.g, col.b};
	}

	static void process_mesh(mesh::Mesh3D& mesh, aiMesh* assimpMesh, const aiScene* scene, const std::vector<assets::AssetID>& assets) {
		assert( assimpMesh != nullptr );
		assert( scene != nullptr );

		// process vertex data
		for ( auto idx = 0U; idx < assimpMesh->mNumVertices; ++idx ) {
			mesh::Vertex3D vertex{
				.position = convert(assimpMesh->mVertices[idx]),
				.normal = convert(assimpMesh->mNormals[idx])
			};

			if ( assimpMesh->mTextureCoords[0] != nullptr ) {
				vertex.texPos = convert2( assimpMesh->mTextureCoords[0][idx] );
			}

			mesh.data.push_back(vertex);
		}

		// process faces (indexes)
		for ( auto idx = 0U; idx < assimpMesh->mNumFaces; ++idx ) {
			auto face = assimpMesh->mFaces[idx];
			assert( face.mNumIndices == 3);

			for ( auto vid = 0U; vid < face.mNumIndices; ++vid) {
				mesh.indices.push_back( face.mIndices[vid] );
			}
		}

		// process material
		if ( assimpMesh->mMaterialIndex >= 0) {
			mesh.material = assets[assimpMesh->mMaterialIndex];
		}
	}

	static void process_node(mesh::MultiMesh3D& mesh, aiNode* node, const aiScene* scene, const std::vector<assets::AssetID>& assets) {
		for ( auto idx = 0U; idx < node->mNumMeshes; ++idx ) {
			auto *assimpMesh = scene->mMeshes[ node->mMeshes[idx] ];

			// process assimp submesh
			mesh::Mesh3D meshData;
			process_mesh(meshData, assimpMesh, scene, assets);

			mesh.meshes.push_back(meshData);
		}

		for ( auto idx = 0U; idx < node->mNumChildren; ++idx) {
			process_node(mesh, node->mChildren[idx], scene, assets);
		}
	}

	struct AssetStuff {
		assets::Loader* loader;
		assets::AssetManager* manager;

		inline void checkLoaded(assets::AssetID asset) {
			if (!manager->has(asset)) {
				debug::info("triggered JIT upload for {}, use the chain loader", asset);
				loader->load(asset);
			}
		}

		inline auto isLoaded(assets::AssetID asset) const -> bool {
			return manager->has(asset);
		}

		inline void set(assets::AssetID asset, auto* assetPtr) {
			assert(assetPtr != nullptr);
			manager->set(asset, assetPtr);
		}

	};

	static auto process_texture( const aiTextureType type, const aiMaterial* assimpMat, AssetStuff& stuff, const assets::LoaderContext& config) -> std::vector<assets::AssetID> {
		std::vector<assets::AssetID> matRefs;
		matRefs.reserve( assimpMat->GetTextureCount(type) );

		// iterate through things
		for ( auto i = 0U ; i < assimpMat->GetTextureCount(type); ++i ) {
			aiString texName;
			assimpMat->GetTexture( type, i, &texName );
			const auto texID = config.makeRef(texName.C_Str());

			// trigger asset loading
			stuff.checkLoaded( texID );

			// assets
			matRefs.push_back( texID );
		}
		return matRefs;
	}

	static auto process_mat_colour(const aiMaterial* mat, const char* name, int a1, int a2) -> math::vec3 {
		aiColor3D col{0.0F, 0.0F, 0.0F};
		mat->Get( name, a1, a2, col );
		debug::info("read colour: {}, {}, {}, {}", name, col.r, col.g, col.b);

		return convert(col);
	}

	static void process_material(const assets::AssetID& guid, aiMaterial* assimpMat, AssetStuff stuff, const assets::LoaderContext& config) {
		auto* material = new mesh::Material();

		debug::info("processing: {}", guid);

		// for each material, calculate what it's name would be then request it
		material->diffuseTextures = process_texture( aiTextureType_DIFFUSE, assimpMat, stuff, config );
		material->normalTextures = process_texture( aiTextureType_NORMALS, assimpMat, stuff, config );
		material->specularTextures = process_texture( aiTextureType_SPECULAR, assimpMat, stuff, config );

		material->ambient = process_mat_colour( assimpMat, AI_MATKEY_COLOR_AMBIENT );
		material->diffuse = process_mat_colour( assimpMat, AI_MATKEY_COLOR_DIFFUSE );
		material->specular = process_mat_colour( assimpMat, AI_MATKEY_COLOR_SPECULAR );

		stuff.set( guid, material );
	}

	auto load_assimp_model(assets::Loader* loader, const assets::AssetID& guid, const assets::LoaderContext& data, void* userPtr) -> assets::AssetRefRaw {
//		auto *filePath = std::get<assets::AssetPath *>(data);
		auto filePath = data.assetPath;

		AssetStuff stuff {
			.loader = loader,
			.manager = (assets::AssetManager*)userPtr
		};

		if ( stuff.isLoaded(guid) ) {
			// asset in DB, what do you want me to do?
			return nullptr;
		}

		// assimp stuff
		Assimp::Importer importer;

		// import the scene from disk
		const aiScene *scene = importer.ReadFile(filePath.c_str(), aiProcessPreset_TargetRealtime_Fast | aiProcess_FlipUVs);
		if ( scene == nullptr || (scene->mFlags & AI_SCENE_FLAGS_INCOMPLETE) != 0 || scene->mRootNode == nullptr ) {
			debug::warning("unable to load required model asset");
			return nullptr;
		}

		debug::debug("Processing assimp mesh, {} meshes, {} materials, {} textures", scene->mNumMeshes, scene->mNumMaterials, scene->mNumTextures);

		// calculate the mapping from material => asset mappings
		auto parentRel = std::filesystem::relative( filePath.parent_path(), filePath.parent_path().parent_path() );
		std::vector<assets::AssetID> matAssets;
		for ( auto idx = 0U; idx < scene->mNumMaterials; ++idx ) {
			aiString matName = scene->mMaterials[idx]->GetName();
			auto matRel = parentRel / matName.C_Str();
			matAssets.push_back( assets::make_asset_id_rt( data.pack, matRel ) );
		}

		// now we can try importing the mesh data
		auto* packedMesh = new mesh::MultiMesh3D();
		process_node( *packedMesh, scene->mRootNode, scene, matAssets);

		// now we try importing the materials
		for (auto idx = 0U; idx < scene->mNumMaterials; ++idx ) {
			process_material( matAssets[idx], scene->mMaterials[idx], stuff, data );
		}

		// FIXME asset loading system needs rework, this is bonkers.
		stuff.set(guid, packedMesh);
		return nullptr;
	}

	auto load_assimp_texture(assets::Loader* /*loader*/, const assets::AssetID& guid, const assets::LoaderContext& data, void* userPtr) -> assets::AssetRefRaw {
		auto* manager = (assets::AssetManager*)userPtr;

		if ( manager->has(guid) ) {
			// already loaded.
			return nullptr;
		}

		auto filePath = data.assetPath;
		stbi_set_flip_vertically_on_load(1);

		//load the texture data into memory
		auto* image = new data::Texture2D();
		image->data = stbi_load(filePath.c_str(), &image->size.x, &image->size.y, &image->channels, 3);

		if ( image->data == nullptr ) {
			debug::warning("error reading texture: {}", stbi_failure_reason());
			return nullptr;
		} else {
			debug::info("image reports it loaded correctly, adding {} to database", guid);
			manager->set(guid, image );
		}

		return nullptr;
	}

	auto load_tex_stb(const std::filesystem::path& filePath, assets::MemoryBlock& block) -> bool {
		stbi_set_flip_vertically_on_load(1);

		//load the texture data into memory
		auto* image = new data::Texture2D();
		image->data = stbi_load(filePath.c_str(), &image->size.x, &image->size.y, &image->channels, 3);

		if ( image->data == nullptr ) {
			debug::warning("error reading texture: {}", stbi_failure_reason());
			delete image;
			return false;
		} else {
			// TODO pass metadata to loader in a sensible way
			block.size = image->channels * image->size.x * image->size.y;
			block.data = (std::byte*)image->data;
			delete image;
			return true;
		}
	}

	auto is_tex_stb(const std::filesystem::path& filePath) -> assets::AssetTypeID {
		// detect jpgs
		if ( filePath.extension() == ".jpg" || filePath.extension() == ".jpeg" ) {
			return TEXTURE_RGBA;
		}

		// detect png
		if ( filePath.extension() == ".png" ) {
			return TEXTURE_RGBA;
		}

		return assets::INVALID_ASSET_TYPE;
	}

	auto is_model_assimp(const std::filesystem::path& filePath) -> assets::AssetTypeID {
		auto ext = filePath.extension().string();
		std::transform(ext.begin(), ext.end(), ext.begin(), ::tolower);

		if ( ext == ".obj" || ext == ".fbx" ) {
			return MODEL_MULTI3D;
		}

		return assets::INVALID_ASSET_TYPE;
	}

	auto extract_requirements(const std::string& packName, const std::filesystem::path& packRoot, assets::ResourceRecord& rr) -> bool {
		Assimp::Importer importer;
		const aiScene *scene = importer.ReadFile( rr.m_path, aiProcess_Triangulate | aiProcess_FlipUVs);
		if ( !scene || scene->mFlags & AI_SCENE_FLAGS_INCOMPLETE || !scene->mRootNode ) {
			debug::warning("unable to load required model asset");
			return false;
		}

		// we want to find out about dependencies
		auto baseDir = std::filesystem::relative( rr.m_path.parent_path(), packRoot );
		for ( auto idx = 0U; idx < scene->mNumMaterials; ++idx) {
			const auto *mat = scene->mMaterials[idx];

			std::array matTypes { aiTextureType_DIFFUSE, aiTextureType_NORMALS, aiTextureType_SPECULAR };
			for ( auto& matType : matTypes ) {
				const auto diffCount = mat->GetTextureCount(matType);
				for (auto idx2 = 0U; idx2 < diffCount; ++idx2) {
					aiString texName;
					mat->GetTexture(matType, idx2, &texName);

					auto assetPath = baseDir / texName.C_Str();
					auto assetRequired = assets::make_asset_id_rt(packName, assetPath);
					rr.m_requires.push_back(assetRequired);
				}
			}

		}

		return true;
	}

	auto AssimpModule::factory(modules::ServiceName service, modules::Services &serviceManager) -> bool {
		if ( service == MODEL_PROVIDER ) {
			auto* assetLoader = serviceManager.get<assets::Loader>();
			assetLoader->setFactory( MODEL_MULTI3D, load_assimp_model, assets::LoadType::PATH );
			assetLoader->setFactory( TEXTURE_RGBA, load_assimp_texture, assets::LoadType::PATH );

			// new loading system
			auto* checkin = serviceManager.get<assets::CheckinAdapted>();
			checkin->setLoader( MIME_JPG, load_tex_stb, is_tex_stb );
			checkin->setLoader( MIME_PNG, load_tex_stb, is_tex_stb );

			checkin->setLoader( MIME_OBJ, assets::NEEDS_CHECKIN, is_model_assimp );
			checkin->setLoader( MIME_FBX, assets::NEEDS_CHECKIN, is_model_assimp );
			checkin->setProcessor( MIME_OBJ, extract_requirements);
			checkin->setProcessor( MIME_FBX, extract_requirements );
			return false;
		}
		return false;
	}

} // namespace fggl::data
